# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

MIN_ENSEMBL_RELEASE = 48
MAX_ENSEMBL_RELEASE = 78

def check_release_number(release):
    """
    Convert a user-provided release number into
    an integer, check to make sure it's in the
    valid range of Ensembl releases
    """
    try:
        release = int(release)
    except:
        raise ValueError("Invalid Ensembl release: %s" % release)

    if release < MIN_ENSEMBL_RELEASE or release > MAX_ENSEMBL_RELEASE:
        raise ValueError(
            "Invalid Ensembl releases %d, must be between %d and %d" % (
                release, MIN_ENSEMBL_RELEASE, MAX_ENSEMBL_RELEASE))
    return release

# mapping from Ensembl release to which reference assembly it uses
_human_references = {}

# Ensembl release 48-54 use NCBI36 as a reference
for i in range(48, 54 + 1):
    _human_references[i] = 'NCBI36'

# Ensembl releases 55-75 use GRCh37 as a reference
for i in range(55, 75 + 1):
    _human_references[i] = 'GRCh37'

# Most recent Ensembl releases use GRCh38
for i in range(76, MAX_ENSEMBL_RELEASE + 1):
    _human_references[i] = 'GRCh38'

def which_human_reference_name(release):
    release = check_release_number(release)
    if release not in _human_references:
        raise ValueError(
            "No reference found for release %d" % release)
    return _human_references[release]



"""
Test all methods which return collections of gene IDs that aren't converting
from some other tyoe of name or ID.

TODO: Implement tests for EnsemblRelease.gene_ids
"""
from __future__ import absolute_import
from .test_common import test_ensembl_releases
from pyensembl import EnsemblRelease

def test_gene_ids_ensembl77_hla_a():
    # chr6:29,945,884  is a position for HLA-A
    # Gene ID = ENSG00000206503
    # based on:
    # http://useast.ensembl.org/Homo_sapiens/Gene/
    # Summary?db=core;g=ENSG00000206503;r=6:29941260-29945884
    ids = (EnsemblRelease(77, auto_download=True).
           gene_ids_at_locus(6, 29945884))
    expected = "ENSG00000206503"
    assert ids == ["ENSG00000206503"], \
        "Expected HLA-A, gene ID = %s, got: %s" % (expected, ids)

@test_ensembl_releases()
def test_gene_ids_on_contig(ensembl):
    gene_ids_chr17 = ensembl.gene_ids(contig=17)
    # gene ID of TP53
    tp53 = "ENSG00000141510"
    assert tp53 in gene_ids_chr17, \
        "Missing %s from %s on chr17, example IDs: %s (total = %d)" % (
            tp53, ensembl, gene_ids_chr17[:5], len(gene_ids_chr17))

    # gene ID of SMAD4
    gene_ids_chr18 = ensembl.gene_ids(contig=18)
    smad4 = "ENSG00000141646"
    assert smad4 in gene_ids_chr18, \
        "Missing %s from %s on chr18, example result: %s (total = %d)" % (
            smad4, ensembl, gene_ids_chr18[:5], len(gene_ids_chr18))




# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from typechecks import require_integer, require_string
from memoized_property import memoized_property

from .common import memoize
from .locus import Locus

class Exon(Locus):
    def __init__(self, exon_id, db):
        require_string(exon_id, "exon ID")

        self.id = exon_id
        self.db = db

        columns = [
            'seqname',
            'start',
            'end',
            'strand',
            'gene_name',
            'gene_id',
        ]

        result = self.db.query_one(
            select_column_names=columns,
            filter_column='exon_id',
            filter_value=exon_id,
            feature='exon',
            distinct=True)

        result_dict = {}
        for i, column_name in enumerate(columns):
            result_dict[column_name] = result[i]

        Locus.__init__(
            self,
            result_dict['seqname'],
            result_dict['start'],
            result_dict['end'],
            result_dict['strand'])

        self.gene_name = result_dict['gene_name']
        self.gene_id = result_dict['gene_id']

    def __str__(self):
        return "Exon(exon_id=%s, gene_name=%s, contig=%s, start=%d, end=%s)" % (
            self.id, self.gene_name, self.contig, self.start, self.end)

    def __repr__(self):
        return str(self)

    # possible annotations associated with exons
    _EXON_FEATURES = {'start_codon', 'stop_codon', 'UTR', 'CDS'}

    @memoize
    def _exon_feature_positions(self, feature):
        """
        Find start and end positions of features (such as start codons)
        which are contained within this exon.
        """
        if feature not in self._EXON_FEATURES:
            raise ValueError("Invalid exon feature: %s" % feature)

        # query for distinct ranges since, for example, multiple transcripts
        # often have the same start codon. Each transcript's start codon has
        # its own feature='start_codon' entry
        query = """
            SELECT DISTINCT start, end
            FROM %s
            WHERE seqname = ?
            AND strand = ?
            AND start >= ?
            AND end <= ?
        """ % (feature,)

        query_params = [
            self.contig,
            self.strand,
            self.start,
            self.end,
        ]
        cursor = self.db.connection.execute(query, query_params)
        results = cursor.fetchall()

        # check to make sure we only got back integer values
        for (start, end) in results:
            require_integer(start, "start position")
            require_integer(end, "end position")
        return results

    @memoized_property
    def start_codon_positions(self):
        """
        Absolute positions of overlapping start codons.
        """
        return self._exon_feature_positions('start_codon')

    @memoized_property
    def stop_codon_positions(self):
        """
        Absolute positions of overlapping stop codons.
        """
        return self._exon_feature_positions('stop_codon')

    def _first_offset(self, start, end):
        relative_start, relative_end = self.offset_range(start, end)
        return min(relative_start, relative_end)

    @memoize
    def _exon_feature_offsets(self, feature):
        """
        Start and end offsets (relative to this exon) of features such as
        start_codon and stop_codon.
        """
        # start and end positions on the chromosome
        absolute_positions = self._exon_feature_positions(feature)
        results = []
        for start, end in absolute_positions:
            local_position = self._first_offset(start, end)
            results.append(local_position)
        return results

    @memoized_property
    def start_codon_offsets(self):
        """
        How many bases from the beginning of the exon (starting from 0)
        is the first base of the start codon?
        """
        return self._exon_feature_offsets('start_codon')

    @memoized_property
    def stop_codon_offsets(self):
        """
        How many bases from the beginning of the exon (starting from 0)
        is the first base of the stop codon?
        """
        return self._exon_feature_offsets('stop_codon')

    @memoized_property
    def contains_start_codon(self):
        """
        Does this exon contain a start codon in any transcript?
        """
        return len(self.start_codon_positions) > 0

    @memoized_property
    def contains_stop_codon(self):
        """
        Does this exon contain a stop codon in any transcript?
        """
        return len(self.stop_codon_positions) > 0



from __future__ import absolute_import

from .test_common import major_releases

from nose.tools import nottest

@nottest
def check_id_length(feature_name):
    for release in major_releases:
        # only load chromosome Y to speed up tests
        df = release.gtf.dataframe(contig="Y")
        assert feature_name in df, \
            "%s not a column in DataFrame for %s" % (feature_name, release)

        # Ensembl IDs are formatted like ENSG00000223972
        # which is always length 15
        ids = df[feature_name].dropna()

        assert len(ids) > 0, "No %s found" % feature_name

        # any entry missing this ID should be dropped
        id_lengths = ids.str.len()
        correct_mask = (id_lengths == 15) | (id_lengths == 0)
        assert correct_mask.all(), \
           "Invalid %s IDs: %s" % (
                feature_name,
                ids[~correct_mask])

def test_gene_id_length():
    check_id_length('gene_id')

def test_transcript_id_length():
    check_id_length('transcript_id')

def test_protein_id_length():
    check_id_length('protein_id')



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from typechecks import is_string

CACHE_SUBDIR = "ensembl"

def memoize(fn):
    """Simple memoization decorator for functions and methods,
    assumes that all arguments to the function can be hashed and
    compared.
    """
    memoized_values = {}

    @wraps(fn)
    def wrapped_fn(*args, **kwargs):
        key = (args, tuple(sorted(kwargs.items())))
        if key not in memoized_values:
            memoized_values[key] = fn(*args, **kwargs)
        return memoized_values[key]

    return wrapped_fn

def is_valid_ensembl_id(ensembl_id):
    """Is the argument a valid ID for any Ensembl feature?"""
    return is_string(ensembl_id) and ensembl_id.startswith("ENS")

def require_ensembl_id(ensembl_id):
    if not is_valid_ensembl_id(ensembl_id):
        raise ValueError("Invalid Ensembl ID '%s'" % ensembl_id)

def is_valid_human_transcript_id(transcript_id):
    """Is the argument a valid identifier for human Ensembl transcripts?"""
    return is_string(transcript_id) and transcript_id.startswith("ENST")

def require_human_transcript_id(transcript_id):
    if not is_valid_human_transcript_id(transcript_id):
        raise ValueError("Invalid transcript ID '%s'" % transcript_id)

def is_valid_human_protein_id(protein_id):
    """Is the argument a valid identifier for human Ensembl proteins?"""
    return is_string(protein_id) and protein_id.startswith("ENSP")

def require_human_protein_id(protein_id):
    if not is_valid_human_protein_id(protein_id):
        raise ValueError("Invalid protein ID '%s'" % protein_id)



"""
Test all methods which return collections of gene names that aren't converting
from some other type of name or ID.
"""
from __future__ import absolute_import
from pyensembl import EnsemblRelease
from .test_common import test_ensembl_releases

# make sure that familia
KNOWN_GENE_NAMES = [
    "TP53",
    "ERBB2",
    "SMAD4",
    "CTAG1A",
]

@test_ensembl_releases()
def test_all_gene_names(ensembl):
    """
    test_all_gene_names : Make sure some known gene names such as
    SMAD4, HSP90AA1, TP53, ERBB2
    """
    gene_names = ensembl.gene_names()
    print(type(gene_names))
    for gene_name in KNOWN_GENE_NAMES:
        print(str(gene_name) in set(gene_names))
        assert gene_name in gene_names, \
            "Missing gene name %s from %s" % (gene_name, ensembl)

def test_gene_names_at_locus_ensembl77_hla_a():
    # chr6:29,945,884  is a position for HLA-A
    # based on:
    # http://useast.ensembl.org/Homo_sapiens/Gene/
    # Summary?db=core;g=ENSG00000206503;r=6:29941260-29945884
    names = (EnsemblRelease(77, auto_download=True).
             gene_names_at_locus(6, 29945884))
    assert names == ["HLA-A"], "Expected gene name HLA-A, got: %s" % (names,)

@test_ensembl_releases()
def test_gene_names_on_contig(ensembl):
    gene_names_chr17 = ensembl.gene_names(17)
    assert "TP53" in gene_names_chr17, \
        "No TP53 in gene names on chr17 of %s, gene names: %s ... (%d)" % (
            ensembl, list(gene_names_chr17[:4]), len(gene_names_chr17))

    gene_names_chr18 = ensembl.gene_names(18)
    assert "SMAD4" in gene_names_chr18, \
        "No SMAD4 in gene names on chr18 of %s, gene names: %s ... (%d)" % (
            ensembl, list(gene_names_chr18[:4]), len(gene_names_chr18))



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_latin_to_common_names = {
    "ailuropoda_melanoleuca" : ["panda"],
    "anas_platyrhynchos" : ["mallard", "duck"],
    "anolis_carolinensis" : ["carolina anole"],
    "astyanax_mexicanus" : ["mexican tetra"],
    "bos_taurus" : ["cattle", "cow"],
    "caenorhabditis_elegans" : ["c. elegans", "nematode"],
    "callithrix_jacchus" : ["marmoset"],
    "canis_familiaris" : ["dog"],
    "cavia_porcellus" : ["guinea pig"],
    "chlorocebus_sabaeus" : ["green monkey", "sabaeus monkey"],
    "choloepus_hoffmanni" : ["hoffman's two-toed sloth"],
    "ciona_intestinalis" : ["vase tunicate"],
    "ciona_savignyi" : [
        "pacific transparent sea squirt",
        "solitary sea squirt"],
    "danio_rerio" : ["zebrafish"],
    "dasypus_novemcinctus" : ["nine-banded armadillo"],
    "dipodomys_ordii" : ["ord's kangaroo rat"],
    "drosophila_melanogaster" : [
        "fruit fly",
        "vinegar fly"],
    "echinops_telfairi" : ["lesser hedgehog tenrec"],
    "equus_caballus" : ["horse"],
    "erinaceus_europaeus" : [
        "european hedgehog",
        "hedgehog"],
    "felis_catus" : ["cat", "domestic cat"],
    "ficedula_albicollis" : ["collared flycatcher"],
    "gadus_morhua" : ["atlnatic cod"],
    "gallus_gallus" : ["chicken", "domesticated fowl"],
    "gasterosteus_aculeatus" : ["three-spined stickleback"],
    "gorilla_gorilla" : ["gorilla"],
    "homo_sapiens" : ["human"],
    "ictidomys_tridecemlineatus" : [
        "thirteen-lined ground squirrel",
        "striped gopher",
        "leopard ground squirrel"
        "squinney"],
    "latimeria_chalumnae" : [
        "west indian ocean coelacanth",
        "african coelacanth"],
    "lepisosteus_oculatus" : ["spotted gar"],
    "loxodonta_africana" : ["african bush elephant"],
    "macaca_mulatta" : ["rhesus macaque"],
    "macropus_eugenii" : [
        "tammar wallaby",
        "dama wallaby",
        "darma wallaby"],
    "meleagris_gallopavo" : ["wild turkey"],
    "microcebus_murinus" : ["gray mouse lemur"],
    "monodelphis_domestica" : ["gray short-tailed opossum"],
    "mus_musculus" : ["house mouse", ],
    "mustela_putorius_furo" : ["domestic ferret"],
    "myotis_lucifugus" : ["little brown bat"],
    "nomascus_leucogenys" : ["northern white-cheeked gibbon"],
    "ochotona_princeps" : ["american pika"],
    "oreochromis_niloticus" : ["nile tilapia"],
    "ornithorhynchus_anatinus" : ["platypus"],
    "oryctolagus_cuniculus" : ["european rabbit", "rabbit"],
    "oryzias_latipes" : ["japanese rice fish", "medaka"],
    "otolemur_garnettii" : [
        "northern greater galago",
        "small-eared greater galago"],
    "ovis_aries" : ["sheep"],
    "pan_troglodytes" : ["chimpanzee"],
    "papio_anubis" : ["olive baboon"],
    "pelodiscus_sinensis" : ["chinese softshell turtle"],
    "petromyzon_marinus" : ["sea lamprey"],
    "poecilia_formosa" : ["amazon molly"],
    "pongo_abelii" : ["sumatran orangutan"],
    "procavia_capensis" : ["rock hyrax"],
    "pteropus_vampyrus" : [
        "large flying fox",
        "large fruit bat",
        "kalang",
        "kalong"],
    "rattus_norvegicus" : ["brown rat", "rat", "norwegian rat"],
    "saccharomyces_cerevisiae" : ["yeast"],
    "sarcophilus_harrisii" : ["tasmanian devil"],
    "sorex_araneus" : ["shrew"],
    "sus_scrofa" : ["wild boar"],
    "taeniopygia_guttata" : ["zebra finch"],
    "takifugu_rubripes" : ["japanese puffer", "tiger puffer", "torafugu"],
    "tarsius_syrichta" : ["philippine tarsier", "mawmag", "mamag"],
    "tetraodon_nigroviridis" : ["green spotted puffer"],
    "tupaia_belangeri" : ["northern treeshrew"],
    "tursiops_truncatus" : ["bottlenose dolphin"],
    "vicugna_pacos" : ["whale"],
    "xenopus_tropicalis" : ["western clawed frog"],
    "xiphophorus_maculatus" : ["southern platyfish"],
}

_common_to_latin_name = {}
for (latin_name, common_names) in _latin_to_common_names.items():
    for common_name in common_names:
        assert common_name not in _common_to_latin_name, \
            "Common name %s (for %s) appears twice" % (common_name, latin_name)
        _common_to_latin_name[common_name] = latin_name

def normalize_species_name(name):
    name = name.lower()

    # if name is already in form like "homo_sapiens" then just return it
    if name in _latin_to_common_names:
        return name

    # if species name was "homo sapiens" then replace spaces with underscores
    # and return "homo_sapiens"
    if " " in name:
        no_spaces = name.replace(" ", "_")
        if no_spaces in _latin_to_common_names:
            return no_spaces

    # if given a common name such as "human", look up its latin equivalent
    if name in _common_to_latin_name:
        return _common_to_latin_name[name]

    raise ValueError("Unknown species name: %s" % name)




from __future__ import absolute_import

from pyensembl import EnsemblRelease

from nose.tools import assert_raises, ok_

ensembl = EnsemblRelease(77, auto_download=True)

def test_gene_ids_of_gene_name_hla_release77():
    hla_a_gene_ids = ensembl.gene_ids_of_gene_name("HLA-A")
    assert 'ENSG00000206503' in hla_a_gene_ids, hla_a_gene_ids

    hla_b_gene_ids = ensembl.gene_ids_of_gene_name("HLA-B")
    assert 'ENSG00000234745' in hla_b_gene_ids, hla_b_gene_ids

    hla_c_gene_ids = ensembl.gene_ids_of_gene_name("HLA-C")
    assert 'ENSG00000204525' in hla_c_gene_ids, hla_c_gene_ids

def test_gene_id_of_protein_id_release77():
    gene_id = ensembl.gene_id_of_protein_id("ENSP00000485677")
    ok_('ENSG00000279634', gene_id)

def test_gene_id_of_invalid_name():
    with assert_raises(Exception, None):
        ensembl.gene_ids_of_gene_name("A wonderous pony sees through your soul")



from __future__ import absolute_import

from pyensembl import EnsemblRelease

import pandas as pd
from nose.tools import eq_

ensembl = EnsemblRelease(75, auto_download=True)

def test_dataframe_column_at_locus():
    series = ensembl.gtf.dataframe_column_at_locus(
        column_name='gene_id',
        contig='1',
        start=11869)
    assert series is not None
    assert len(series) > 0
    eq_('ENSG00000223972', series.unique()[0])

def test_dataframe_at_locus():
    df = ensembl.gtf.dataframe_at_locus(contig='1',
                                        start=11869)
    assert df is not None
    assert len(df) > 0
    eq_('ENSG00000223972', df['gene_id'].unique()[0])

def test_slice_column():
    column_name_series = pd.Series(["a", "b", "c"])
    start_series = pd.Series([1000, 2000, 3000])
    end_series = pd.Series([4000, 2300, 3500])

    series = ensembl.gtf._slice_column(
        column_name_series, start_series, end_series,
        900, 1000)
    eq_({'a'}, set(series))

    series = ensembl.gtf._slice_column(
        column_name_series, start_series, end_series,
        1000, 5000)
    eq_({'a', 'b', 'c'}, set(series))

    series = ensembl.gtf._slice_column(
        column_name_series, start_series, end_series,
        1900, 2000)
    eq_({'a', 'b'}, set(series))

    series = ensembl.gtf._slice_column(
        column_name_series, start_series, end_series,
        1900, 3000)
    eq_({'a', 'b', 'c'}, set(series))

    series = ensembl.gtf._slice_column(
        column_name_series, start_series, end_series,
        3000, 6000)
    eq_({'a', 'c'}, set(series))



"""
Check that pyensembl returns correct exon objects for exon IDs
and loci. Make sure the information on the exon object matches
the expected gene ID and location.
"""
from __future__ import absolute_import

from pyensembl import EnsemblRelease

ensembl77 = EnsemblRelease(77, auto_download=True)

def test_exon_object_by_id():
    """
    test_exon_object_by_id : check properties of exon 4 of CTNNB1 when looked
    up by ID in Ensembl 77.
    """
    exon = ensembl77.exon_by_id("ENSE00003464041")
    assert exon.gene_name == "CTNNB1", \
        "Unexpected gene name: %s" % exon.gene_name
    assert exon.contig == "3", exon.contig
    assert exon.strand == "+"
    assert exon.on_forward_strand
    assert exon.on_positive_strand
    assert exon.start ==  41224526, "Unexpected exon start: %s" % exon.start
    assert exon.end == 41224753, "Unexpected exon end: %s" % exon.end
    assert exon.length ==  228

def test_exon_object_by_id_on_negative_strand():
    """
    test_exon_object_by_id : check properties of exon 1 from CXCR3 when looked
    up by ID in Ensembl 77.
    """
    exon = ensembl77.exon_by_id("ENSE00001817013")
    assert exon.gene_name == "CXCR3", \
        "Unexpected gene name: %s" % exon.gene_name
    assert exon.contig == "X", exon.contig
    assert exon.strand == "-"
    assert exon.on_backward_strand
    assert exon.on_negative_strand
    assert exon.start ==  71618438, "Unexpected exon start: %s" % exon.start
    assert exon.end == 71618517, "Unexpected exon end: %s" % exon.end
    assert exon.length ==  80


def test_exon_object_at_locus():
    """
    test_exon_object_at_locus : check properties of exon 4 of CTNNB1 when looked
    up by its location on the forward strand of chr3
    """
    exons = ensembl77.exons_at_locus(3, 41224526, strand="+")
    for exon in exons:
        assert exon.gene_name == "CTNNB1", exon.transcript_name
        assert exon.contig == "3", exon.contig
        assert exon.strand == "+"
        assert exon.on_forward_strand
        assert exon.on_positive_strand
        assert exon.start <=  41224526, "Unexpected exon start: %s" % exon.start
        assert exon.end >= 41224526, "Unexpected exon end: %s" % exon.end

def test_exon_object_at_locus_on_negative_strand():
    """
    test_exon_object_at_locus : check properties of exon 1 of CXCR3 when looked
    up by its location on the negative strand of chrX
    """
    exons = ensembl77.exons_at_locus("chrX", 71618517, strand="-")
    for exon in exons:
        assert exon.gene_name == "CXCR3", exon.transcript_name
        assert exon.contig == "X", exon.contig
        assert exon.strand == "-"
        assert exon.on_backward_strand
        assert exon.on_negative_strand
        assert exon.start <=  71618517, "Unexpected exon start: %s" % exon.start
        assert exon.end >= 71618517, "Unexpected exon end: %s" % exon.end


def test_contains_start_codon():
    """
    test_contains_start_codon : Test that first exon of CXCR3-002 contains
    a start codon.
    """
    transcript = ensembl77.transcripts_by_name("CXCR3-002")[0]
    exon = transcript.exons[0]
    assert exon.contains_start_codon

def test_contains_stop_codon():
    """
    Test that second exon of CXCR3-002 contains a stop codon.
    """
    transcript = ensembl77.transcripts_by_name("CXCR3-002")[0]
    exon = transcript.exons[1]
    assert exon.contains_stop_codon

def test_partial_start_codons():
    """
    test_partial_start_codons : Make sure none of the start codons returned
    only partially overlap with an exon.
    """
    # Exon ENSE00003718948 has a start codon immediately before the
    # exon's start position on the forward strand
    exon = ensembl77.exon_by_id("ENSE00003718948")
    for (start, end) in exon.start_codon_positions:
        assert start >= exon.start, \
            "Exon at locus [%d, %d], start codon at %d" % (
                exon.start, exon.end, start)

    for offset in exon.start_codon_offsets:
        assert offset >= 0, \
            "Invalid negative offset for start codon: %d" % offset
        assert offset != 0, \
            "Partially overlapping start codon should not be included"

def test_start_codon_offset():
    """
    test_start_codon_offset : Ensure that start codon in exon #1 of CXCR3-002
    is 68 nucleotides from the start of the exon.
    """
    transcript = ensembl77.transcripts_by_name("CXCR3-002")[0]
    exon = transcript.exons[0]
    start_offsets = exon.start_codon_offsets
    assert len(start_offsets) == 1, \
        "First exon of CXCR3-002 should only overlap one start codon"
    offset = start_offsets[0]
    assert offset == 68, \
        "Expected stop codon 68nt from start of exon #1, got %s" % (offset,)

def test_stop_codon_offset():
    """
    test_stop_codon_offset : Ensure that the only stop codon
    overlapping exon #2 of CXCR-002 is 1092 nucleotides
    from the start of the exon.
    """
    transcript = ensembl77.transcripts_by_name("CXCR3-002")[0]
    exon = transcript.exons[1]
    stop_offsets = exon.stop_codon_offsets
    assert len(stop_offsets) == 1, \
        "Last exon of CXCR3-002 should only overlap one stop codon, got: %s" % (
        stop_offsets)
    offset = stop_offsets[0]
    assert offset == 1092, \
        "Expected stop codon 1092nt from start of exon #2, got %s" % (offset,)




from __future__ import absolute_import

import functools

from pyensembl import EnsemblRelease
from nose.tools import nottest


_cached_releases = {}

def cached_release(version):
    if version in _cached_releases:
        return _cached_releases[version]
    ensembl = EnsemblRelease(version, auto_download=True)
    _cached_releases[version] = ensembl
    return ensembl

ensembl_grch36 = cached_release(54)  # last release for GRCh36/hg18
ensembl_grch37 = cached_release(75)  # last release for GRCh37/hg19
ensembl_grch38 = cached_release(78)  # most recent release for GRCh38

major_releases = [
    ensembl_grch37,
    ensembl_grch38
]

contigs = list(range(1, 23)) + ["X", "Y", "M"]

@nottest
def test_ensembl_releases(*versions):
    """
    Run a unit test which takes an EnsemblRelease as an argument
    for multiple releases (most recent for each reference genome)
    """
    if len(versions) == 0:
        ensembl_releases = major_releases
    else:
        ensembl_releases = [cached_release(version) for version in versions]

    def decorator(test_fn):
        @functools.wraps(test_fn)
        def new_test_fn():
            for ensembl in ensembl_releases:
                test_fn(ensembl)
        return new_test_fn
    return decorator



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Cache and serializing the results of expensive computations. Used in pyensembl
primarily to cache the heavy-weight parsing of GTF files and various
filtering operations on Ensembl entries.

A piece of data is returned from one of three sources:
1) Cache cold. Run the user-supplied compute_fn.
2) Cache warm on disk. Parse or unpickle the serialized result into memory.
3) Cache warm in memory. Return cached object.
"""
from __future__ import print_function, division, absolute_import

try:
    # Python 2: cPickle faster than pickle.
    import cPickle as pickle
except ImportError:
    # Python 3
    import pickle

import logging
from os import remove, stat
from os.path import exists

import pandas as pd

_memory_cache = {}

def is_empty(filename):
    return stat(filename).st_size == 0

def delete_file(path):
    if exists(path):
        logging.info("Deleting cached file %s" % path)
        remove(path)

def remove_from_cache(key):
    if key in _memory_cache:
        del _memory_cache[key]
        delete_file(key)

def clear_cached_objects():
    for key in _memory_cache.keys():
        delete_file(key)
    _memory_cache.clear()

def _read_csv(csv_path):
    print("Reading Dataframe from %s" % csv_path)
    df = pd.read_csv(csv_path)
    if 'seqname' in df:
        # by default, Pandas will infer the type as int,
        # then switch to str when it hits non-numerical
        # chromosomes. Make sure whole column has the same type
        df['seqname'] = df['seqname'].map(str)
    return df

def _write_csv(df, csv_path):
    print("Saving DataFrame to %s" % csv_path)
    df.to_csv(csv_path, index=False)

def cached_dataframe(csv_path, compute_fn):
    """
    If a CSV path is in the _memory_cache, then return that cached value.

    If we've already saved the DataFrame as a CSV then load it.

    Otherwise run the provided `compute_fn`, and store its result
    in memory and and save it as a CSV.
    """
    if not csv_path.endswith(".csv"):
        raise ValueError("Invalid path '%s', must be a CSV file" % csv_path)

    if csv_path in _memory_cache:
        return _memory_cache[csv_path]

    if exists(csv_path) and not is_empty(csv_path):
        df = _read_csv(csv_path)
    else:
        df = compute_fn()
        if not isinstance(df, pd.DataFrame):
            raise TypeError(
                "Expected compute_fn to return DataFrame, got %s : %s" % (
                    df, type(df)))
        _write_csv(df, csv_path)
    _memory_cache[csv_path] = df
    return df

def cached_object(path, compute_fn):
    """
    If `cached_object` has already been called for a value of `path` in this
    running Python instance, then it should have a cached value in the
     _memory_cache; return that value.

    If this function was never called before with a particular value of
    `path`, then call compute_fn, and pickle it to `path`.

    If `path` already exists, unpickle it and store that value in
    _memory_cache.
    """
    if path in _memory_cache:
        return _memory_cache[path]

    if exists(path) and not is_empty(path):
        with open(path, 'rb') as f:
            obj = pickle.load(f)
    else:
        obj = compute_fn()
        with open(path, 'wb') as f:
            pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
    _memory_cache[path] = obj
    return obj



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import
from os.path import split, join
import pandas as pd

import datacache
from typechecks import require_string

from .gtf_parsing import load_gtf_as_dataframe
from .common import CACHE_SUBDIR
from .locus import normalize_chromosome, normalize_strand
from .compute_cache import cached_dataframe, clear_cached_objects
from .url_templates import ENSEMBL_FTP_SERVER, gtf_url


class GTF(object):
    """
    Download and parse a GTF gene annotation file from a given URL.
    Represent its contents as a Pandas DataFrame (optionally filtered
    by locus, column, contig, &c).
    """
    def __init__(
            self,
            release,
            species="homo_sapiens",
            server=ENSEMBL_FTP_SERVER,
            auto_download=False,
            decompress=True):
        self.cache = datacache.Cache(CACHE_SUBDIR)
        self.release = release
        self.decompress = decompress
        self.url = gtf_url(
            ensembl_release=release,
            species=species,
            server=server)

        self.remote_filename = split(self.url)[1]

        assert self.remote_filename.endswith(".gtf.gz"), \
            "Expected remote GTF file %s to end with '.gtf.gz'" % (
                self.remote_filename,)

        self.auto_download = auto_download

        # lazily load DataFrame of all GTF entries if necessary
        # for database construction
        self._dataframes = {}

    def clear_cache(self):
        # clear any dataframes we constructed
        self._dataframes.clear()

        # clear cached dataframes loaded from CSV
        clear_cached_objects()

    def base_filename(self):
        """
        Trim extensions such as ".gtf" or ".gtf.gz",
        leaving only the base filename which should be used
        to construct other derived filenames for cached data.
        """
        assert ".gtf" in self.remote_filename, \
            "GTF filename must contain .gtf extension: %s" % (
                self.remote_filename,)

        return self.remote_filename.split(".gtf")[0]

    def local_filename(self):
        """Filename used for local copy of GTF"""
        if self.decompress:
            return self.base_filename() + ".gtf"
        else:
            return self.remote_filename

    def local_copy_exists(self):
        """Has a local copy of the GTF file been downloaded?"""
        return self.cache.exists(
            self.url,
            self.local_filename(),
            decompress=self.decompress)

    def local_gtf_path(self):
        """
        Returns local path to GTF file for given release of Ensembl,
        download from the Ensembl FTP server if not already cached and
        auto download is enabled.
        """
        if self.local_copy_exists() or self.auto_download:
            # Does a download if the cache is empty
            return self.cache.fetch(
                self.url,
                self.local_filename(),
                decompress=self.decompress)
        raise ValueError('Ensembl annotation data is not currently '
                         'installed for release %s. Run '
                         '"pyensembl install --release %s" or call '
                         '"EnsemblRelease(%s).install()"' %
                         ((self.release,) * 3))

    def local_dir(self):
        return split(self.local_gtf_path())[0]

    def local_data_file_path(
            self,
            contig=None,
            feature=None,
            column=None,
            strand=None,
            distinct=False,
            extension=".csv"):
        """
        Path to local file for storing materialized views of the Ensembl data.
        Typically this is a CSV file, the filename reflects which filters have
        been applied to the entries of the database.

        Parameters:

        contig : str, optional
            Path for subset of data restricted to given contig

        feature : str, optional
            Path for subset of data restrict to given feature

        column : str, optional
            Restrict to single column

        strand : str, optional
            Positive ("+") or negative ("-") DNA strand. Default = either.

        distinct : bool, optional
            Only keep unique values (default=False)
        """
        base = self.local_filename()
        dirpath = self.local_dir()
        csv_filename = base + ".expanded"
        if contig:
            contig = normalize_chromosome(contig)
            csv_filename += ".contig.%s" % (contig,)
        if feature:
            csv_filename += ".feature.%s" % (feature,)
        if column:
            csv_filename += ".column.%s" % (column,)
        if strand:
            if strand == "+":
                strand_string = "positive"
            elif strand == "-":
                strand_string = "negative"
            else:
                raise ValueError("Invalid strand value: %s" % strand)
            csv_filename += ".strand.%s" % strand_string
        if distinct:
            csv_filename += ".distinct"
        csv_filename += extension
        return join(dirpath, csv_filename)

    def _load_full_dataframe(self):
        """
        Loads full dataframe from cached CSV or constructs it from GTF
        """
        csv_path = self.local_data_file_path()
        return cached_dataframe(csv_path, self._load_full_dataframe_from_gtf)

    def _load_full_dataframe_from_gtf(self):
        """
        Parse this release's GTF file and load it as a Pandas DataFrame
        """
        path = self.local_gtf_path()
        return load_gtf_as_dataframe(path)

    def dataframe(self, contig=None, feature=None, strand=None):
        """
        Load Ensembl entries as a DataFrame, optionally restricted to
        particular contig or feature type.
        """
        if contig:
            contig = normalize_chromosome(contig)

        if strand:
            strand = normalize_strand(strand)

        if feature is not None:
            require_string(feature, "feature")

        key = (contig, feature, strand)

        if key not in self._dataframes:
            csv_path = self.local_data_file_path(
                contig=contig,
                feature=feature,
                strand=strand,
                distinct=False)

            def local_loader_fn():
                # pylint: disable=no-member
                # pylint has trouble with df.seqname and similar
                # statements in this function.

                full_df = self._load_full_dataframe()
                assert len(full_df) > 0, \
                    "Dataframe representation of Ensembl database empty!"

                # rename since we're going to be filtering the entries but
                # may still want to access the full dataset
                df = full_df
                if contig:
                    df = df[df.seqname == contig]
                    if len(df) == 0:
                        raise ValueError("Contig not found: %s" % (contig,))

                if feature:
                    df = df[df.feature == feature]
                    if len(df) == 0:
                        # check to make sure feature was somewhere in
                        # the full dataset before returning an empty dataframe
                        features = full_df.feature.unique()
                        if feature not in features:
                            raise ValueError(
                                "Feature not found: %s" % (feature,))
                if strand:
                    df = df[df.strand == strand]

                return df

            self._dataframes[key] = cached_dataframe(csv_path, local_loader_fn)

        return self._dataframes[key]

    def dataframe_column_at_locus(
            self,
            column_name,
            contig,
            start,
            end=None,
            offset=None,
            strand=None):
        """
        Subset of entries which overlap an inclusive range of loci
        """
        if end is None and offset is None:
            end = start
        elif offset is None:
            end = start + offset - 1

        df_contig = self.dataframe(contig=contig, strand=strand)

        assert column_name in df_contig, \
            "Unknown Ensembl property: %s" % column_name

        return self._slice_column(
            df_contig[column_name],
            df_contig.start,
            df_contig.end,
            start,
            end)

    @staticmethod
    def _slice_column(column_name_series, start_series, end_series,
                      start, end):
        """
        Given a Series of data, and two Series' representing start
        and end positions for the end, keep all data for which
        the start and end positions overlap the start and end
        parameters (inclusively).

        For example:
        column_name_series: ENSG00000223972, ENSG00000000003, etc.
        start_series: 11869, 100635558, etc.
        end_series: 14409, 100635252, etc.
        start = 100000000
        end = 110000000

        Returns only ENSG00000000003, since it overlaps
        [100000000, 110000000].
        """
        df = pd.DataFrame({'name': column_name_series,
                           'start': start_series,
                           'end': end_series})
        df = GTF._slice(df, 'start', 'end', start, end)
        return df.name

    @staticmethod
    def _slice(df, df_start_col_name, df_end_col_name, start, end):
        """
        Given a DataFrame, along with the names of columns in the
        DataFrame representing start and end positions, keep all
        data for which the start and end positions overlap the start
        and end parameters (inclusively).
        """
        # No overlap because the whole thing is before start
        mask_left = df[df_end_col_name] < start
        # No overlap because the whole thing is after end
        mask_right = df[df_start_col_name] > end
        df = df[~mask_left & ~mask_right]
        return df

    def dataframe_at_locus(
            self,
            contig,
            start,
            end=None,
            offset=None,
            strand=None):
        """
        Subset of entries which overlap an inclusive range of
        chromosomal positions
        """
        if end is None and offset is None:
            end = start
        elif offset is None:
            end = start + offset - 1

        df_contig = self.dataframe(contig=contig, strand=strand)

        # find genes whose start/end boundaries overlap with the position
        return GTF._slice(df_contig, df_contig.start.name,
                          df_contig.end.name, start, end)

    def download(self, force=False):
        """
        Download the GTF file if one does not exist. If `force` is
        True, overwrites any existing file.

        Returns True if a download happened.
        """
        if not force and self.local_copy_exists():
            return False

        self.cache.fetch(
            self.url,
            self.local_filename(),
            decompress=self.decompress,
            force=force)
        return True



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from memoized_property import memoized_property
from typechecks import require_string, require_instance

from .biotypes import is_valid_biotype
from .database import Database
from .locus import Locus
from .transcript import Transcript

class Gene(Locus):

    def __init__(self, gene_id, ensembl):
        require_string(gene_id, "gene ID")
        self.id = gene_id
        # can't check the type of ensembl since it will create a circular
        # dependency between this module and ensembl_release but note that
        # it's expected to be an EnsemblRelease object
        self.ensembl = ensembl
        self.db = ensembl.db
        require_instance(self.db, Database, "db")
        columns = [
            'gene_name',
            'seqname',
            'start',
            'end',
            'strand',
            'gene_biotype'
        ]
        gene_name, contig, start, end, strand, biotype = self.db.query_one(
            columns,
            filter_column='gene_id',
            filter_value=gene_id,
            feature='gene')
        if not gene_name:
            raise ValueError("Missing name for gene with ID = %s" % gene_id)
        self.name = gene_name

        Locus.__init__(self, contig, start, end, strand)

        if not is_valid_biotype(biotype):
            raise ValueError(
                "Invalid gene_biotype %s for gene with ID = %s" % (
                    biotype, gene_id))
        elif biotype:
            self.biotype = biotype
        else:
            raise ValueError(
                "Missing gene_biotype for gene with ID = %s" % gene_id)

    def __str__(self):
        return "Gene(id=%s, name=%s, location=%s:%d-%d)" % (
            self.id,
            self.name,
            self.contig,
            self.start,
            self.end)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
            isinstance(other, Gene) and
            self.id == other.id and
            self.ensembl == other.ensembl)

    def __hash__(self):
        return hash(self.id)

    @memoized_property
    def transcripts(self):
        """
        Property which dynamically construct transcript objects for all
        transcript IDs associated with this gene.
        """
        transcript_id_results = self.db.query(
            select_column_names=['transcript_id'],
            filter_column='gene_id',
            filter_value=self.id,
            feature='transcript',
            distinct=False,
            required=False)

        # We're doing a SQL query for each transcript ID to fetch
        # its particular information, might be more efficient if we
        # just get all the columns here, but how do we keep that modular?
        return [
            Transcript(result[0], self.ensembl)
            for result in transcript_id_results
        ]

    @memoized_property
    def exons(self):
        exons_dict = {}
        for transcript in self.transcripts:
            for exon in transcript.exons:
                if exon.id not in exons_dict:
                    exons_dict[exon.id] = exon
        return list(exons_dict.values())



from __future__ import absolute_import

import tempfile

from pyensembl import compute_cache

import pandas as pd
from nose.tools import raises

class Counter(object):
    """
    Use this class to count how many times a function gets called by
    cached_object and cached_dataframe.
    """
    def __init__(self):
        self.count = 0

    def increment(self):
        self.count += 1
        return self.count

    def increment_dataframe(self):
        value = self.increment()
        return pd.DataFrame({'x': [value]})

def test_cached_object_with_tempfile():
    """
    test_cached_object_with_tempfile : A temporary file exists before
    calling into compute_cache.cached_object but is empty, should be treated
    as if result has never been computed before (rather than trying to load
    the empty file).
    """
    counter = Counter()
    with tempfile.NamedTemporaryFile() as f:
        # call repeatedly to test the hot and cold cache logic
        result = compute_cache.cached_object(
            f.name, compute_fn=counter.increment)
        assert result == 1, "Expected result=1, got %s" % (result,)
        assert counter.count == 1, \
            "Expected compute_fn to be called once, got %s" % (counter.count,)


def test_cached_dataframe_with_tempfile():
    """
    test_cached_dataframe_with_tempfile : A temporary file exists before
    calling into compute_cache.cached_dataframe but is empty,
    should be treated as if result has never been computed before
    (rather than trying to load the empty file).
    """
    counter = Counter()
    with tempfile.NamedTemporaryFile(suffix='.csv') as f:
        # call repeatedly to test hot and cold cache logic
        for _ in range(2):
            df = compute_cache.cached_dataframe(
                f.name, compute_fn=counter.increment_dataframe)
            # get counter value from inside of dataframe
            result = df['x'].ix[0]
            assert result == 1, \
                "Expected result=1, got %s" % (result,)
            assert counter.count == 1, \
                "Expected compute_fn to be called once, got %s" % (
                    counter.count,)

def test_cached_dataframe_returns_correct_type():
    def make_a_dataframe():
        return pd.DataFrame({'x': [0, 1, 2]})
    with tempfile.NamedTemporaryFile(suffix='.csv') as f:
        # call repeatedly to test the cold and hot cache logic
        for _ in range(2):
            df = compute_cache.cached_dataframe(
                f.name, compute_fn=make_a_dataframe)
            assert isinstance(df, pd.DataFrame), \
                "Expected DataFrame, got %s : %s" % (df, type(df))

def test_cached_object_with_list_returns_correct_type():
    def make_a_list():
        return [1, 2, 3]
    with tempfile.NamedTemporaryFile() as f:
        # call repeatedly to test the cold and hot cache logic
        for _ in range(2):
            df = compute_cache.cached_object(
                f.name, compute_fn=make_a_list)
            assert isinstance(df, list), \
                "Expected list, got %s : %s" % (df, type(df))

@raises(Exception)
def test_dataframe_path_must_be_csv():
    # compute_cache should raise an exception when filename doesn't
    # end with .csv extension
    compute_cache.cached_dataframe(
        csv_path="tempfile_not_csv",
        compute_fn=lambda _: pd.DataFrame({'x': []}))



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from immuno.mutate import mutate_protein_from_transcript
import immuno.ensembl.annotation as ensembl
from immuno.ensembl.transcript_data import EnsemblReferenceData

# Test case CTNNB1
# Transcript Id: ENST00000453024
# Gene Id : ENSG00000168036
# 17 Exons
# Length 2,841
# Chrom : 3
# Start : 41,240,936
# End :  41,281,227

ref_data = EnsemblReferenceData()


def test_get_strand_CASP9():
    genomic_transcript = "ENST00000333868"
    forward = ensembl.is_forward_strand(genomic_transcript)
    assert not forward

def test_get_strand_CTNNB1():
    genomic_transcript = "ENST00000453024"
    forward = ensembl.is_forward_strand(genomic_transcript)
    assert forward

def test_load_CTNNB1_cdna_transcript():
    genomic_transcript = "ENST00000453024"
    transcript = ref_data.get_cdna(genomic_transcript)
    assert transcript is not None
    assert len(transcript) == 2841, (transcript, len(transcript))


def test_load_CTNNB1_protein_transcript():
    protein_transcript = "ENSP00000427553"
    transcript = ref_data.get_protein(protein_transcript)
    assert transcript is not None
    assert transcript[0] == 'M', (transcript, len(transcript))
    assert transcript[-1] == 'E', (transcript, len(transcript))

def test_load_CTNNB1_exon_from_transcript():
    transcript_id = "ENST00000453024"
    exons = ensembl.get_exons_from_transcript(transcript_id)
    assert exons.shape[0] == 17

    transcript_id = "ENST00000405570"
    exons = ensembl.get_exons_from_transcript(transcript_id)
    assert exons.shape[0] == 16

def test_load_CTNNB1_exon_from_transcript_length():

    transcript_id = 'ENST00000405570'
    transcript = ref_data.get_cdna(transcript_id)
    exons = ensembl.get_exons_from_transcript(transcript_id)
    exons['length'] = exons['seq_region_end_exon'] - exons['seq_region_start_exon'] + 1
    assert exons['length'].sum() == len(transcript), exons

def test_load_SMAD4_cdna_transcript():
    transcript_id = "ENST00000342988"
    transcript = ref_data.get_cdna(transcript_id)
    assert transcript is not None
    assert len(transcript) == 8769, len(transcript)
    assert transcript[0] == 'A', transcript[0]
    assert transcript[-1] == 'T', transcript[-1]

def test_get_gene_from_pos():
    variant = {
        'chr': '3',
        'pos': 41250936,
        'ref': 'A',
        'alt': 'C'
    }
    vcf = pd.DataFrame.from_records([variant])
    transcripts_df = ensembl.annotate_vcf_transcripts(vcf)
    assert "ENSG00000168036" in set(transcripts_df['stable_id_gene'])

def test_get_transcript_from_pos():
    variant = {
        'chr': '3',
        'pos': 41250936,
        'ref': 'A',
        'alt': 'C'
    }
    vcf = pd.DataFrame.from_records([variant])
    transcripts_ids = ensembl.annotate_vcf_transcripts(vcf)
    assert "ENST00000453024" in set(transcripts_ids['stable_id_transcript'])

def test_get_all_transcript_from_pos():
    variant = {
        'chr': '3',
        'pos': 41275636,
        'ref': 'G',
        'alt': 'A'
    }
    vcf = pd.DataFrame.from_records([variant])
    transcripts_ids = ensembl.annotate_vcf_transcripts(vcf)
    transcript_ids = set(transcripts_ids['stable_id_transcript'])
    assert "ENST00000405570" in transcript_ids
    assert "ENST00000396183" in transcript_ids
    assert "ENST00000349496" in transcript_ids
    assert "ENST00000453024" in transcript_ids
    assert "ENST00000396185" in transcript_ids

def test_get_transcript_index_from_pos():
    variant = {
        'chr': '3',
        'pos': 41275636,
        'ref': 'G',
        'alt': 'A'
    }
    transcript_id = 'ENST00000405570'
    idx = ensembl.get_transcript_index_from_pos(
        41275636, transcript_id, skip_untranslated_region=False)
    assert idx == 1686, idx

    transcript = ref_data.get_cdna(transcript_id)
    assert transcript[idx] == variant['ref']

def test_get_5prime_utr_length_RET():

    transcript_id = "ENST00000355710"
    exons = ensembl.get_exons_from_transcript(transcript_id)

    utr_length = ensembl.get_five_prime_utr_length(exons)
    assert utr_length == 232

def test_get_3prime_utr_length_RET():
    transcript_id = "ENST00000355710"
    exons = ensembl.get_exons_from_transcript(transcript_id)
    utr_length = ensembl.get_three_prime_utr_length(exons)
    assert utr_length == 2082

def test_get_5prime_utr_length_CTNNB1():
    transcript_id = "ENST00000405570"
    exons = ensembl.get_exons_from_transcript(transcript_id)
    utr_length = ensembl.get_five_prime_utr_length(exons)
    assert utr_length == 156

def test_get_3prime_utr_length_CTNNB1():
    transcript_id = "ENST00000405570"
    exons = ensembl.get_exons_from_transcript(transcript_id)
    utr_length = ensembl.get_three_prime_utr_length(exons)
    assert utr_length == 11

def test_get_5prime_utr_length_reverse_strand_CASP9():

    transcript_id = "ENST00000333868"
    exons = ensembl.get_exons_from_transcript(transcript_id)

    utr_length = ensembl.get_five_prime_utr_length(exons, forward=False)
    assert utr_length == 95

def test_get_3prime_utr_length_reverse_strand_CASP9():
    transcript_id = "ENST00000333868"
    exons = ensembl.get_exons_from_transcript(transcript_id)
    utr_length = ensembl.get_three_prime_utr_length(exons, forward=False)
    assert utr_length == 673

def test_get_transcript_and_mutate_vcf():
    variant = {
        'chr': '10',
        'pos': 43617416,
        'ref': 'T',
        'alt': 'C'
    }

    vcf = pd.DataFrame.from_records([variant])
    transcripts_ids = ensembl.annotate_vcf_transcripts(vcf)

    transcript_ids = set(transcripts_ids['stable_id_transcript'])
    assert "ENST00000355710" in transcript_ids
    assert "ENST00000340058" in transcript_ids
    transcript_id = "ENST00000355710"

    cdna_idx = ensembl.get_transcript_index_from_pos(
        variant['pos'], transcript_id, skip_untranslated_region=False)
    assert cdna_idx is not None
    assert cdna_idx < 5569
    cdna_transcript = ref_data.get_cdna(transcript_id)
    assert(cdna_transcript[cdna_idx] == variant['ref'])

    cds_idx = ensembl.get_transcript_index_from_pos(
        variant['pos'], transcript_id, skip_untranslated_region=True)
    assert cds_idx is not None
    cds_transcript = ref_data.get_cds(transcript_id)
    assert(cds_transcript[cds_idx] == variant['ref'])

    region = mutate_protein_from_transcript(
            cds_transcript,
            cds_idx,
            variant['ref'],
            variant['alt'],
            padding=10)
    assert region is not None
    assert len(region.seq) == 21, (region.seq, len(region.seq))
    assert region.seq == 'RSQGRIPVKWTAIESLFDHIY'


def test_peptide_from_transcript_PARS2():
    transcript_id = 'ENST00000371279'
    variant = {
        'chr': '1',
        'pos': 55224569,
        'ref': 'T',
        'alt': 'G'
    }

    cds_idx = ensembl.get_transcript_index_from_pos(
        variant['pos'], transcript_id, skip_untranslated_region=True)
    assert cds_idx is not None
    assert cds_idx == 265


if __name__ == '__main__':
    from dsltools import testing_helpers
    testing_helpers.run_local_tests()



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from typechecks import is_integer, require_string

def normalize_chromosome(c):
    if is_integer(c):
        if c == 0:
            raise ValueError("Contig cannot be 0")
        c = str(c)

    require_string(c, "contig name", nonempty=True)

    # only strip off lowercase chr since some of the non-chromosomal contigs
    # start with "CHR"
    if c.startswith("chr"):
        c = c[3:]

    # standardize mitochondrial genome to be "MT"
    if c == "M":
        return "MT"
    # just in case someone is being lazy, capitalize "X" and "Y"
    elif c == "x":
        return "X"
    elif c == "y":
        return "Y"
    else:
        return c

def normalize_strand(strand):
    if strand == "+" or strand == "-":
        return strand
    elif strand == 1:
        return "+"
    elif strand == -1:
        return "-"

    raise ValueError("Invalid strand: %s" % (strand,))

class Locus(object):
    """
    Base class for any entity which can be localized at a range of positions
    on a particular strand of a chromosome/contig.
    """

    def __init__(
            self,
            contig,
            start,
            end,
            strand):
        """
        contig : str
            Chromosome or other sequence name in the reference assembly

        start : int
            Start position of locus on the contig

        end : int
            Inclusive end position on the contig

        strand : str
            Should we read the locus forwards ('+') or backwards ('-')?
        """

        self.contig = normalize_chromosome(contig)
        self.strand = normalize_strand(strand)

        start = int(start)
        end = int(end)

        if start == 0:
            raise ValueError("Expected start > 0 (using base 1 coordinates)")
        elif end == 0:
            raise ValueError("Expected end > 0 (using base 1 coordinates)")

        if end < start:
            raise ValueError(
                "Expected start <= end, got start = %d, end = %d" % (
                    start, end))
        self.start = start
        self.end = end

    def __str__(self):
        return "Locus(contig=%s, start=%s, end=%s, strand=%s)" % (
            self.contig, self.start, self.end, self.strand)

    def __repr__(self):
        return str(self)

    def __len__(self):
        return self.end - self.start + 1

    def __eq__(self, other):
        return (
            isinstance(other, Locus) and
            self.contig == other.contig and
            self.start == other.start and
            self.end == other.end and
            self.strand == other.strand
        )

    def __hash__(self):
        return hash(str(self))

    @property
    def length(self):
        return self.end - self.start + 1

    def position_offset(self, position):
        if position > self.end or position < self.start:
            raise ValueError(
                "Position %d outside valid range %d..%d of %s" % (
                    position, self.start, self.end, self))
        elif self.on_forward_strand:
            return position - self.start
        else:
            return self.end - position

    def offset_range(self, start, end):
        """
        Database start/end entries are always ordered such that
        start < end. This makes computing a relative position (e.g. of a stop
        codon relative to its transcript) complicated since the "end"
        position of a backwards locus is actually earlir on the strand.
        This function correctly selects a start vs. end value depending
        on this locuses's strand and determines that position's offset from
        the earliest position in this locus.
        """
        assert start <= end, \
            "Locations should always have start < end, got start=%d, end=%d" % (
                start, end)

        if start < self.start or end > self.end:
            raise ValueError("Range (%d, %d) falls outside %s" % (
                start, end, self))

        if self.on_forward_strand:
            return (start - self.start, end - self.start)

        else:
            return (self.end - end, self.end - start)

    def on_contig(self, contig):
        return normalize_chromosome(contig) == self.contig

    def on_strand(self, strand):
        return normalize_strand(strand) == self.strand

    @property
    def on_forward_strand(self):
        return self.on_strand("+")

    @property
    def on_positive_strand(self):
        return self.on_forward_strand

    @property
    def on_backward_strand(self):
        return self.on_strand("-")

    @property
    def on_negative_strand(self):
        return self.on_backward_strand

    def can_overlap(self, contig, strand=None):
        """
        Is this locus on the same contig and (optionally) on the same strand?
        """
        return (
            self.on_contig(contig)
            and
            (strand is None or self.on_strand(strand)))

    def distance_to_interval(self, start, end):
        """
        Find the distance between intervals [start1, end1] and [start2, end2].
        If the intervals overlap then the distance is 0.
        """
        if self.start > end:
            # interval is before this exon
            return self.start - end
        elif self.end < start:
            # exon is before the interval
            return start - self.end
        else:
            return 0

    def distance_to_locus(self, other):
        if not self.can_overlap(other.contig, other.strand):
            # if two loci are on different contigs or strands,
            # can't compute a distance between them
            return float("inf")
        return self.distance_to_interval(other.start, other.end)

    def overlaps(self, contig, start, end, strand=None):
        """
        Does this locus overlap with a given range of positions?

        Since locus position ranges are inclusive, we should make sure
        that e.g. chr1:10-10 overlaps with chr1:10-10
        """
        return (
            self.can_overlap(contig, strand)
            and
            self.distance_to_interval(start, end) == 0)

    def overlaps_locus(self, other_locus):
        return self.overlaps(
            other_locus.contig,
            other_locus.start,
            other_locus.end,
            other_locus.strand)

    def contains(self, contig, start, end, strand=None):
        return (
            self.can_overlap(contig, strand)
            and
            start >= self.start
            and
            end <= self.end)

    def contains_locus(self, other_locus):
        return self.contains(
            other_locus.contig,
            other_locus.start,
            other_locus.end,
            other_locus.strand)



from __future__ import absolute_import

from mock import Mock, patch
from nose.tools import assert_raises

from pyensembl import EnsemblRelease

# pylint: disable=no-value-for-parameter
# pylint and mocking don't go very well together. pylint complains,
# for example, about _test_db_index not being called with the
# mock_index parameter.

@patch('pyensembl.sequence_data.SequenceData.index')
def _test_db_index(mock_index, db_exists):
    """
    Return True if the GTF database gets created, which should
    be different depending on whether the database already existed.

    Note: we need to mock the reference transcript indexing, as we're
    testing GTF indexing.
    """
    data = EnsemblRelease(54)
    data.db._connect_if_exists = Mock(return_value=db_exists)
    data.db._create_database = Mock()
    data.index(force=False)

    return data.db._create_database.called

def test_db_not_exists_index():
    called = _test_db_index(db_exists=False)
    assert called, "Expected a new database"

def test_db_exists_index():
    called = _test_db_index(db_exists=True)
    assert not called, "Expected no new database"

def test_auto_download_off():
    data = EnsemblRelease(54)
    data.db._connect_if_exists = Mock(return_value=False)
    assert_raises(ValueError, data.gene_names_at_locus,
        contig=6, position=29945884)



"""Make sure we're getting correct transcritp sequence from Ensembl and that
it's a sequence type which correctly implements `complement`
and `reverse_complement`
"""

from __future__ import absolute_import
from nose.tools import eq_
from pyensembl import EnsemblRelease


ensembl54 = EnsemblRelease(54, auto_download=True)

def test_transcript_sequence_ensembl54():

    seq = ensembl54.transcript_sequence("ENST00000321606")
    assert len(seq) == 414, \
        "Expected transcript ENST00000321606 to have 414nt, got %s : %d" % (
            seq, len(seq))
    nucleotide_lines = [
        "CATGTCACCCACCTTCAGGCGGCCCAAGACACTGCGACTCCGGAGGCAGCCCAGATATCCTCGGAAGAG",
        "CACCCCCAGGAGAAACAAGCTTGGCCACTATGCTATCATCAAGTTTCCGCTGACCACTGAGTCGGCCGT",
        "GAAGAAGATAGAAGAAAACAACACGCTTGTGTTCACTGTGGATGTTAAAGCCAACAAGCACCAGATCAG",
        "ACAGGCTGTGAAGAAGCTCTATGACAGTGATGTGGCCAAGGTCACCACCCTGATTTGTCCTGATAAAGA",
        "GAACAAGGCATATGTTCGACTTGCTCCTGATTATGATGCTTTCGATGTTGTAACAAAATTGGGATCACC",
        "TAAACTGAGTCCAGCTGGCTAACTCTAAATATATGTGTATCTTTTCAGCATAAAAAAATAATGTTTTTC"
    ]
    full_transcript_sequence = "".join(nucleotide_lines)
    eq_(str(seq), full_transcript_sequence)

    # make sure the Sequence object's reverse and complement properties work
    reverse_complement = seq.complement()
    reverse_complement_lines = [
        "GAAAAACATTATTTTTTTATGCTGAAAAGATACACATATATTTAGAGTTAGCCAGCTGGACTCAGTTTA",
        "GGTGATCCCAATTTTGTTACAACATCGAAAGCATCATAATCAGGAGCAAGTCGAACATATGCCTTGTTC",
        "TCTTTATCAGGACAAATCAGGGTGGTGACCTTGGCCACATCACTGTCATAGAGCTTCTTCACAGCCTGT",
        "CTGATCTGGTGCTTGTTGGCTTTAACATCCACAGTGAACACAAGCGTGTTGTTTTCTTCTATCTTCTTC",
        "ACGGCCGACTCAGTGGTCAGCGGAAACTTGATGATAGCATAGTGGCCAAGCTTGTTTCTCCTGGGGGTG",
        "CTCTTCCGAGGATATCTGGGCTGCCTCCGGAGTCGCAGTGTCTTGGGCCGCCTGAAGGTGGGTGACATG"
    ]
    reverse_complement = "".join(reverse_complement_lines)
    eq_(len(seq.reverse_complement()), 414)
    eq_(str(seq.reverse_complement()), reverse_complement)



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import
from os import remove
from os.path import exists, split
import logging

from Bio import SeqIO
import datacache

from .common import CACHE_SUBDIR, require_ensembl_id
from .release_info import check_release_number


class SequenceData(object):
    """Container for reference nucleotide and amino acid sequenes.

    Downloads and caches FASTA file, creates database mapping
    identifiers to sequences.
    """
    def __init__(
            self,
            ensembl_release,
            fasta_url,
            auto_download=False):

        # download cache for fetching reference FASTA files
        self.cache = datacache.Cache(CACHE_SUBDIR)
        self.auto_download = auto_download
        self.release = check_release_number(ensembl_release)
        # expect the remote FASTA file to be gzipped and the local needs
        # to be a decompressed text file
        self.fasta_decompress = True
        self.url = fasta_url
        self.remote_filename = split(self.url)[1]

        self.local_filename = self._create_local_filename(self.remote_filename)
        self._init_lazy_fields()

    def _init_lazy_fields(self):
        """Initialize all properties which get lazily constructed on request"""
        # dictionary mapping IDs to sequences
        self._sequence_cache = {}

        # SeqIO.index_db gets lazily constructed since we only want to download
        # the Fasta file if/when we use it
        self._fasta_dictionary = None

        # key set for fasta dictionary, for faster membership tests,
        self._fasta_keys = None

    def _create_local_filename(self, remote_filename):
        """
        We sometimes need to add the release number to a local FASTAfilename
        since some Ensembl releases only have the genome name in the FASTA
        filename but still differ subtly between releases.
        For example, a transcript ID may be missing in Ensembl 75 but present
        in 76, though both have the same FASTA filename
        """
        if ".%d." % self.release in remote_filename:
            return remote_filename

        filename_parts = self.remote_filename.split(".fa.")
        assert len(filename_parts) == 2, \
            "Expected remote filename %s to contain '.fa.gz'" % (
                self.remote_filename,)
        return "".join([
            filename_parts[0],
            ".%d.fa." % self.release,
            filename_parts[1]])

    def __str__(self):
        return "SequenceData(ensembl_release=%d, url=%s, local_path=%s)" % (
            self.release,
            self.url,
            self.local_fasta_path)

    def __repr__(self):
        return str(self)

    def __contains__(self, sequence_id):
        if self._fasta_keys is None:
            self._fasta_keys = set(self.fasta_dictionary.keys())
        return sequence_id in self._fasta_keys

    def __eq__(self, other):
        return (
            isinstance(other, SequenceData) and
            self.url == other.url and
            self.release == other.release)

    @property
    def local_fasta_path(self):
        """
        Returns local path to FASTA file. If it's not already
        cached, download it from the Ensembl FTP server if auto
        download is enabled.
        """
        # If the fasta is already cached, fetching it won't initiate a
        # download. But it's always okay to initiate a download if
        # auto download is enabled.
        if self.local_file_exists() or self.auto_download:
            # Does a download if the cache is empty, otherwise just returns
            # the local path
            return self._fetch(force=False)
        raise ValueError("Ensembl sequence data %s is not currently "
                         "installed for release %s. Run "
                         "\"pyensembl install --relase %s\" or call "
                         "EnsemblRelease(%s).install()" %
                         ((self.local_filename,) + (self.release,) * 3))

    def clear_cache(self):
        """Delete the local FASTA file and its associated database,
        and clear any in-memory cached sequence data.
        """
        # reset in-memory cached object
        self._init_lazy_fields()

        if self.local_file_exists():
            remove(self.local_fasta_path)
        if exists(self.local_database_path):
            remove(self.local_fasta_path)

    @property
    def local_dir(self):
        return split(self.local_fasta_path)[0]

    def local_file_exists(self):
        return self.cache.exists(
                self.url,
                filename=self.local_filename,
                decompress=self.fasta_decompress)

    def _fetch(self, force):
        """Download file from remote URL if not present or if force=True

        Return local path
        """
        return self.cache.fetch(
            url=self.url,
            filename=self.local_filename,
            decompress=self.fasta_decompress,
            force=force)

    def download(self, force=False):
        """Download the FASTA file if one does not exist.

        If `force` is True, overwrites any existing file.
        """
        if not self.local_file_exists() or force:
            self._fetch(force=force)

    @property
    def local_database_path(self):
        return self.local_fasta_path + ".db"

    def _create_or_open_fasta_db(self):
        """Create an index database from a transcript sequence FASTA file"""
        return SeqIO.index_db(
            self.local_database_path,
            self.local_fasta_path,
            "fasta")

    def index(self, force=False):
        """Create a database mapping transcript IDs to sequences.

        Parameters
        -----------
        force : bool
            Recreate database even if it already exists

        Raises an error if the necessary data is not yet downloaded.

        If no error is raised, then after call self._fasta_dictionary
        should be populated.
        """
        # This local_fasta_path property access will raise an error
        # if the necessary data is not yet downloaded
        if exists(self.local_database_path):
            if force:
                logging.info(
                    "Deleting existing sequence database: %s",
                    self.local_database_path)
                remove(self.local_database_path)
            else:
                # try opening the existing FASTA database, if it fails then
                # delete the local file and we'll create a new one further
                # below
                try:
                    self._fasta_dictionary = self._create_or_open_fasta_db()
                    return
                except ValueError as e:
                    # if there was an error opening the database
                    # delete the version we have and try again
                    if "database" in e.message:
                        logging.warn(
                            ("Recreating transcript database %s"
                             " due to error: %s"),
                            self.local_database_path,
                            e.message)
                        remove(self.local_database_path)
                    else:
                        raise
        # if database didn't exist or was incomplete, create it now
        self._fasta_dictionary = self._create_or_open_fasta_db()

    @property
    def fasta_dictionary(self):
        if self._fasta_dictionary is None:
            # should populate self._fasta_dictionary with a database
            # backed dictionary of reference sequences
            self.index()
        return self._fasta_dictionary

    def get(self, sequence_id):
        """Get sequence associated with given ID or return None if missing"""

        if sequence_id not in self._sequence_cache:
            # all Ensembl identifiers start with ENS e.g. ENST, ENSP, ENSE
            require_ensembl_id(sequence_id)

            # get sequence from database
            fasta_record = self.fasta_dictionary.get(sequence_id)
            if fasta_record:
                seq = fasta_record.seq
            else:
                seq = None
            self._sequence_cache[sequence_id] = seq
        return self._sequence_cache[sequence_id]



#!/usr/bin/env python

# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Manipulate pyensembl's local cache.

    %(prog)s {install,download,index} [--release XXX ...]

To install the latest ensembl release:
    %(prog)s install

To install particular release(s):
    %(prog)s install --release 75 77

"""
from __future__ import absolute_import
import argparse
from . import ensembl_release
from .release_info import MAX_ENSEMBL_RELEASE

def run():
    parser = argparse.ArgumentParser(usage=__doc__)
    parser.add_argument('--release',
        type=int,
        nargs="+",
        default=[MAX_ENSEMBL_RELEASE],
        help='Ensembl release. Defaults to latest release %(default)s. '
             'Multiple releases may be specified.')
    parser.add_argument('action', choices=('install', 'download', 'index'),
        help="'install' will download and index any data that is  not "
        "currently downloaded or indexed. 'download' will download data, "
        "regardless of whether it is already downloaded. 'index' will index "
        "data, regardless of whether it is already indexed, and will raise "
        "an error if the data is not already downloaded.")

    args = parser.parse_args()
    for release in args.release:
        ensembl = ensembl_release.EnsemblRelease(release)
        if args.action == 'install':
            ensembl.install()
        elif args.action == 'download':
            ensembl.download()
        elif args.action == 'index':
            ensembl.index()
        else:
            assert False, "shouldn't get here"

if __name__ == '__main__':
    run()



"""
Tests for methods which return collections of transcript IDs that aren't
converting from some type of name or ID.
"""
from __future__ import absolute_import

from pyensembl import EnsemblRelease
from nose.tools import eq_

from .test_common import test_ensembl_releases

# subset of transcript IDs for HLA-A
HLA_A_TRANSCRIPT_IDS = [
        'ENST00000396634',
        'ENST00000376809',
        'ENST00000376806',
        'ENST00000376802',
        'ENST00000496081',
        'ENST00000495183',
        'ENST00000461903',
        'ENST00000479320',
]

def test_transcript_ids_ensembl77_hla_a():
    # chr6:29,945,884  is a position for HLA-A
    # based on:
    # http://useast.ensembl.org/Homo_sapiens/Gene/
    # Summary?db=core;g=ENSG00000206503;r=6:29941260-29945884
    ensembl = EnsemblRelease(77, auto_download=True)
    transcript_ids = ensembl.transcript_ids_at_locus(6, 29941260, 29945884)
    for transcript_id in HLA_A_TRANSCRIPT_IDS:
        assert transcript_id in transcript_ids, \
            "Transcript %s of HLA-A not found overlapping locus" % transcript_id

KNOWN_TRANSCRIPT_IDS = HLA_A_TRANSCRIPT_IDS + [
    'ENST00000398417',  # transcript ID of SMAD4-001
    'ENST00000334701',  # transcript ID of HSP90AA1-001
    'ENST00000599837',  # transcript ID of CTAG1A-002
]

# TODO: add release 54 after transcript IDs for older GTFs are filled in
# See https://github.com/hammerlab/pyensembl/issues/20
@test_ensembl_releases(75, 77)
def test_all_transcript_ids(ensembl):
    transcript_ids = set(ensembl.transcript_ids())
    for transcript_id in KNOWN_TRANSCRIPT_IDS:
        assert transcript_id in transcript_ids, \
            "Missing transcript ID %s from %s" % (transcript_id, ensembl)

@test_ensembl_releases(77)
def test_transcript_id_of_protein_id(ensembl):
    transcript_id = ensembl.transcript_id_of_protein_id(
        "ENSP00000485678")
    eq_("ENST00000623976", transcript_id)



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains the EnsemblRelease class, with its millions of accessors and wrappers
around the Ensembl annotation database.
"""

from __future__ import print_function, division, absolute_import

from glob import glob
import logging
from os.path import join
from os import remove

from .common import require_human_transcript_id, require_human_protein_id
from .compute_cache import cached_object
from .database import Database
from .exon import Exon
from .gene import Gene
from .gtf import GTF
from .release_info import (
    check_release_number,
    MAX_ENSEMBL_RELEASE,
    which_human_reference_name
)
from .sequence_data import SequenceData
from .transcript import Transcript
from .url_templates import ENSEMBL_FTP_SERVER, fasta_url

class EnsemblRelease(object):
    """
    Bundles together the genomic annotation and sequence data associated with
    a particular release of the Ensembl database and provides a wide
    variety of helper methods for accessing this data.
    """

    def __init__(self,
                 release=MAX_ENSEMBL_RELEASE,
                 server=ENSEMBL_FTP_SERVER,
                 auto_download=False):

        self.release = check_release_number(release)
        self.species = "homo_sapiens"
        self.server = server
        self.auto_download = auto_download
        self.reference_name = which_human_reference_name(self.release)

        # GTF object wraps the source GTF file from which we get
        # genome annotations for each Ensembl release. Presents access
        # to each feature annotations as a pandas.DataFrame.
        self.gtf = GTF(
            self.release,
            self.species,
            server,
            auto_download=auto_download)

        # Database object turns the GTF dataframes into sqlite3 tables
        # and wraps them with methods like `query_one`
        self.db = Database(gtf=self.gtf, auto_download=auto_download)

        # get the URL for the cDNA FASTA file containing
        # this releases's transcript sequences
        transcript_sequences_fasta_url = fasta_url(
                ensembl_release=self.release,
                species=self.species,
                sequence_type="cdna",
                server=self.server)
        self.transcript_sequences = SequenceData(
            ensembl_release=self.release,
            fasta_url=transcript_sequences_fasta_url,
            auto_download=auto_download)

        protein_sequences_fasta_url = fasta_url(
            ensembl_release=self.release,
            species=self.species,
            sequence_type="pep",
            server=self.server)
        self.protein_sequences = SequenceData(
            ensembl_release=self.release,
            fasta_url=protein_sequences_fasta_url,
            auto_download=auto_download)

        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)

    def __str__(self):
        return "EnsemblRelease(release=%d, species=%s, genome=%s)" % (
                    self.release,
                    self.species,
                    self.reference_name)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
            isinstance(other, EnsemblRelease) and
            self.release == other.release and
            self.species == other.species)

    def _delete_cached_files(self):
        """
        Any files which start with the same name as our GTF file
        is assumed to be some view of the this release's data and thus
        safe to delete.
        """
        base = self.gtf.base_filename()
        dirpath = self.gtf.local_gtf_path()
        for path in glob(join(dirpath, base + "*")):
            logging.info("Deleting cached file %s", path)
            remove(path)

    def clear_cache(self):
        self.gtf.clear_cache()
        self.transcript_sequences.clear_cache()
        self.protein_sequences.clear_cache()
        self._delete_cached_files()

    def all_feature_values(
            self,
            column,
            feature,
            distinct=True,
            contig=None,
            strand=None):
        """
        Cached lookup of all values for a particular feature property from
        the database, caches repeated queries in memory and
        stores them as a CSV.

        Parameters
        ----------

        column : str
            Name of property (e.g. exon_id)

        feature : str
            Type of entry (e.g. exon)

        distinct : bool, optional
            Keep only unique values

        contig : str, optional
            Restrict query to particular contig

        strand : str, optional
            Restrict results to '+' or '-' strands

        Returns a list constructed from query results.
        """
        # since we're constructing a list, rather than a DataFrame,
        # we're going to store it using pickling rather than the Pandas
        # CSV serializer. Change the default extension from ".csv" to ".pickle"
        pickle_path = self.gtf.local_data_file_path(
            feature=feature,
            column=column,
            contig=contig,
            strand=strand,
            distinct=distinct,
            extension=".pickle")

        def run_query():
            results = self.db.query_feature_values(
                column=column,
                feature=feature,
                distinct=distinct,
                contig=contig,
                strand=strand)
            assert isinstance(results, list), \
                "Expected list from Database.query_feature_values, got %s" % (
                    type(results))
            return results

        return cached_object(pickle_path, compute_fn=run_query)

    def install(self):
        """
        Explicitely download and index any data for this release that
        is not yet downloaded and/or indexed.
        """
        self.download(force=False)
        self.index(force=False)

    def index(self, force=True):
        """
        Create databases and indices for all data for this release

        Parameters
        ----------
        force : bool
            Recreate databases even if already indexed (default = True)

        Raises an error if data is not downloaded.
        """
        self.db.create(force=force)
        self.transcript_sequences.index(force=force)
        self.protein_sequences.index(force=force)

    def transcript_sequence(self, transcript_id):
        """Return cDNA nucleotide sequence of transcript, or None if
        transcript doesn't cDNA sequence.
        """
        require_human_transcript_id(transcript_id)
        return self.transcript_sequences.get(transcript_id)

    def protein_sequence(self, protein_id):
        """Return cDNA nucleotide sequence of transcript, or None if
        transcript doesn't cDNA sequence.
        """
        require_human_protein_id(protein_id)
        return self.transcript_sequences.get(protein_id)

    def download(self, force=True):
        """
        Download all data for this release.

        Parameters
        ----------
        force : bool
            Download data even if we already have a local copy.
        """
        self.gtf.download(force=force)
        self.transcript_sequences.download(force=force)
        self.protein_sequences.download(force=force)

    def genes_at_locus(self, contig, position, end=None, strand=None):
        gene_ids = self.gene_ids_at_locus(
            contig, position, end=end, strand=strand)
        return [self.gene_by_id(gene_id) for gene_id in gene_ids]

    def transcripts_at_locus(self, contig, position, end=None, strand=None):
        transcript_ids = self.transcript_ids_at_locus(
            contig, position, end=end, strand=strand)
        return [
            self.transcript_by_id(transcript_id)
            for transcript_id in transcript_ids
        ]

    def exons_at_locus(self, contig, position, end=None, strand=None):
        exon_ids = self.exon_ids_at_locus(
            contig, position, end=end, strand=strand)
        return [
            self.exon_by_id(exon_id)
            for exon_id in exon_ids
        ]

    def gene_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column='gene_id',
            feature='gene',
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def gene_names_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
             column='gene_name',
             feature='gene',
             contig=contig,
             position=position,
             end=end,
             strand=strand)

    def exon_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column='exon_id',
            feature='exon',
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def transcript_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column='transcript_id',
            feature='transcript',
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def transcript_names_at_locus(
            self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column='transcript_name',
            feature='transcript',
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def protein_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column='protein_id',
            feature='transcript',
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    ###################################################
    #
    #         Methods which return Locus objects
    #         containing (contig, start, stop, strand)
    #         of various genomic entities
    #
    ###################################################

    def locus_of_gene_id(self, gene_id):
        """
        Given a gene ID returns Locus with: chromosome, start, stop, strand
        """
        return self.db.query_locus(
            filter_column='gene_id',
            filter_value=gene_id,
            feature='gene')

    def loci_of_gene_names(self, gene_name):
        """
        Given a gene name returns list of Locus objects with fields:
            chromosome, start, stop, strand
        You can get multiple results since a gene might have multiple copies
        in the genome.
        """
        return self.db.query_loci('gene_name', gene_name, 'gene')

    def locus_of_transcript_id(self, transcript_id):
        return self.db.query_locus(
            filter_column='transcript_id',
            filter_value=transcript_id,
            feature='transcript')

    def locus_of_exon_id(self, exon_id):
        """
        Given an exon ID returns Locus
        """
        return self.db.query_locus('exon_id', exon_id, feature='exon')

    ###################################################
    #
    #             Gene Info Objects
    #
    ###################################################

    def genes(self, contig=None, strand=None):
        """
        Returns all Gene objects in Ensembl. Can be restricted to a
        particular contig/chromosome and strand by the following arguments:

        Parameters
        ----------
        contig : str
            Only return genes on the given contig.

        strand : str
            Only return genes on this strand.
        """
        gene_ids = self.gene_ids(contig=contig, strand=strand)
        return [self.gene_by_id(gene_id) for gene_id in gene_ids]

    def gene_by_id(self, gene_id):
        """
        Construct a Gene object for the given gene ID.
        """
        return Gene(gene_id, ensembl=self)

    def genes_by_name(self, gene_name):
        """
        Get all the unqiue genes with the given name (there might be multiple
        due to copies in the genome), return a list containing a Gene object
        for each distinct ID.
        """
        gene_ids = self.gene_ids_of_gene_name(gene_name)
        return [self.gene_by_id(gene_id) for gene_id in gene_ids]

    def gene_by_protein_id(self, protein_id):
        """
        Get the gene ID associated with the given protein ID,
        return its Gene object
        """
        gene_id = self.gene_id_of_protein_id(protein_id)
        return self.gene_by_id(gene_id)

    ###################################################
    #
    #             Gene Names
    #
    ###################################################

    def _query_gene_name(self, property_name, property_value, feature_type):
        results = self.db.query(
            select_column_names=["gene_name"],
            filter_column=property_name,
            filter_value=property_value,
            feature=feature_type,
            distinct=True,
            required=True)
        return str(results[0][0])

    def gene_names(self, contig=None, strand=None):
        """
        Return all genes in the database,
        optionally restrict to a chromosome and/or strand.
        """
        return self.all_feature_values(
            column='gene_name',
            feature='gene',
            contig=contig,
            strand=strand)

    def gene_name_of_gene_id(self, gene_id):
        return self._query_gene_name("gene_id", gene_id, 'gene')

    def gene_name_of_transcript_id(self, transcript_id):
        return self._query_gene_name(
            "transcript_id", transcript_id, 'transcript')

    def gene_name_of_transcript_name(self, transcript_name):
        return self._query_gene_name(
            "transcript_name", transcript_name, 'transcript')

    def gene_name_of_exon_id(self, exon_id):
        return self._query_gene_name("exon_id", exon_id, 'exon')

    ###################################################
    #
    #             Gene IDs
    #
    ###################################################

    def _query_gene_ids(self, property_name, value, feature='gene'):
        results = self.db.query(
            select_column_names=['gene_id'],
            filter_column=property_name,
            filter_value=value,
            feature=feature,
            distinct=True,
            required=True)
        return [str(result[0]) for result in results if result[0]]

    def gene_ids(self, contig=None, strand=None):
        """
        What are all the gene IDs
        (optionally restrict to a given chromosome/contig and/or strand)
        """
        return self.all_feature_values(
            column='gene_id',
            feature='gene',
            contig=contig,
            strand=strand)

    def gene_ids_of_gene_name(self, gene_name):
        """
        What are the Ensembl gene IDs associated with a given gene name?
        (due to copy events, there might be multiple genes per name)
        """
        results = self._query_gene_ids('gene_name', gene_name)
        if len(results) == 0:
            raise ValueError("Gene name not found: %s" % gene_name)
        return results

    def gene_id_of_protein_id(self, protein_id):
        """
        What is the Ensembl gene ID associated with a given protein ID?
        """
        results = self._query_gene_ids('protein_id', protein_id,
                                       feature='CDS')
        if len(results) == 0:
            raise ValueError("Protein ID not found: %s" % protein_id)
        assert len(results) == 1, \
            ("Should have only one gene ID for a given protein ID, "
             "but found %d: %s" % (len(results), results))
        return results[0]

    ###################################################
    #
    #             Transcript Info Objects
    #
    ###################################################

    def transcripts(self, contig=None, strand=None):
        """
        Construct Transcript object for every transcript entry in
        the Ensembl database. Optionally restrict to a particular
        chromosome using the `contig` argument.
        """
        transcript_ids = self.transcript_ids(contig=contig, strand=strand)
        return [
            Transcript(transcript_id, ensembl=self)
            for transcript_id in transcript_ids
        ]

    def transcript_by_id(self, transcript_id):
        return Transcript(transcript_id, ensembl=self)

    def transcripts_by_name(self, transcript_name):
        transcript_ids = self.transcript_ids_of_transcript_name(transcript_name)
        return [
            self.transcript_by_id(transcript_id)
            for transcript_id in transcript_ids
        ]

    def transcript_by_protein_id(self, protein_id):
        transcript_id = self.transcript_id_of_protein_id(protein_id)
        return self.transcript_by_id(transcript_id)

    ###################################################
    #
    #            Transcript Names
    #
    ###################################################

    def _query_transcript_names(self, property_name, value):
        results = self.db.query(
            select_column_names=['transcript_name'],
            filter_column=property_name,
            filter_value=value,
            feature='transcript',
            distinct=True,
            required=True)
        return [result[0] for result in results]

    def transcript_names(self, contig=None, strand=None):
        """
        What are all the transcript names in the database
        (optionally, restrict to a given chromosome and/or strand)
        """
        return self.all_feature_values(
            column='transcript_name',
            feature='transcript',
            contig=contig,
            strand=strand)

    def transcript_names_of_gene_name(self, gene_name):
        return self._query_transcript_names('gene_name', gene_name)

    def transcript_name_of_transcript_id(self, transcript_id):
        transcript_names = self._query_transcript_names(
            'transcript_id', transcript_id)
        if len(transcript_names) == 0:
            raise ValueError(
                "No transcript names for transcript ID = %s" % transcript_id)
        assert len(transcript_names) == 1, \
            "Multiple transcript names for transcript ID = %s" % transcript_id
        return transcript_names[0]

    ###################################################
    #
    #            Transcript IDs
    #
    ###################################################

    def _query_transcript_ids(self, property_name, value,
                              feature='transcript'):
        results = self.db.query(
            select_column_names=['transcript_id'],
            filter_column=property_name,
            filter_value=value,
            feature=feature,
            distinct=True,
            required=True)
        return [result[0] for result in results]

    def transcript_ids(self, contig=None, strand=None):
        return self.all_feature_values(
            column='transcript_id',
            feature='transcript',
            contig=contig,
            strand=strand)

    def transcript_ids_of_gene_id(self, gene_id):
        return self._query_transcript_ids('gene_id', gene_id)

    def transcript_ids_of_gene_name(self, gene_name):
        return self._query_transcript_ids('gene_name', gene_name)

    def transcript_ids_of_transcript_name(self, transcript_name):
        return self._query_transcript_ids('transcript_name', transcript_name)

    def transcript_ids_of_exon_id(self, exon_id):
        return self._query_transcript_ids('exon_id', exon_id)

    def transcript_id_of_protein_id(self, protein_id):
        """
        What is the Ensembl transcript ID associated with a given
        protein ID?
        """
        results = self._query_transcript_ids('protein_id', protein_id,
                                             feature='CDS')
        if len(results) == 0:
            raise ValueError("Protein ID not found: %s" % protein_id)
        assert len(results) == 1, \
            ("Should have only one transcript ID for a given protein ID, "
             "but found %d: %s" % (len(results), results))
        return results[0]

    ###################################################
    #
    #             Exon Info Objects
    #
    ###################################################

    def exons(self, contig=None, strand=None):
        """
        Create exon object for all exons in the database, optionally
        restrict to a particular chromosome using the `contig` argument.
        """
        # DataFrame with single column called 'exon_id'
        exon_ids = self.exon_ids(contig=contig, strand=strand)
        return [Exon(exon_id, self.db) for exon_id in exon_ids]

    def exon_by_id(self, exon_id):
        return Exon(exon_id, self.db)

    def exon_by_transcript_id_and_number(self, transcript_id, exon_number):
        transcript = self.transcript_by_id(transcript_id)
        if len(transcript.exons) > exon_number:
            raise ValueError(
                "Invalid exon number for transcript %s" % transcript_id)

        # exon numbers in Ensembl are 1-based, need to subtract 1 to get
        # a list index
        return transcript.exons[exon_number - 1]

    ###################################################
    #
    #                Exon IDs
    #
    ###################################################

    def _query_exon_ids(self, property_name, value):
        results = self.db.query(
            select_column_names=['exon_id'],
            filter_column=property_name,
            filter_value=value,
            feature='exon',
            distinct=True,
            required=True)
        return [result[0] for result in results]

    def exon_ids(self, contig=None, strand=None):
        return self.all_feature_values(
            column='exon_id',
            feature='exon',
            contig=contig,
            strand=strand)

    def exon_ids_of_gene_id(self, gene_id):
        return self._query_exon_ids('gene_id', gene_id)

    def exon_ids_of_gene_name(self, gene_name):
        return self._query_exon_ids('gene_name', gene_name)

    def exon_ids_of_transcript_name(self, transcript_name):
        return self._query_exon_ids('transcript_name', transcript_name)

    def exon_ids_of_transcript_id(self, transcript_id):
        return self._query_exon_ids('transcript_id', transcript_id)



from __future__ import absolute_import

from pyensembl import Locus, Transcript
from nose.tools import eq_, assert_not_equal, assert_greater

from .test_common import ensembl_grch38, test_ensembl_releases
from .data import (
    FOXP3_001_transcript_id,
    CTNNBIP1_004_transcript_id,
    CTNNBIP1_004_UTR5,
    CTNNBIP1_004_UTR3,
    CTNNBIP1_004_CDS,
    CTNNBIP1_004_locus,
    CTTNNIP1_004_exon_lengths,
    CTTNNIP1_004_exon_ids,
    EGFR_001_protein_sequence,
)

def test_transcript_start_codon():
    """
    test_transcript_start_codon : Check that fields Transcript
    (for transcript named CTNNBIP1-004) matches known values.
    """
    CTNNBIP1_004_transcript = ensembl_grch38.transcript_by_id(
        CTNNBIP1_004_transcript_id)
    assert Locus.__eq__(CTNNBIP1_004_locus, CTNNBIP1_004_transcript), \
        "Expected locus %s but got %s" % (
            CTNNBIP1_004_locus, Locus.__str__(CTNNBIP1_004_transcript))

    start_offsets = CTNNBIP1_004_transcript.start_codon_spliced_offsets
    assert len(start_offsets) == 3, \
        "Wrong length for start codon: %d (%s)" % (
            len(start_offsets), start_offsets)

    assert all(isinstance(i, int) for i in start_offsets), \
        "Wrong type %s for beginning start codon offset" % (
            [type(i) for i in start_offsets],)

    expected_start_codon_offset = len(CTNNBIP1_004_UTR5)
    start_codon_offset = min(start_offsets)
    assert start_codon_offset == expected_start_codon_offset, \
        "Incorrect start codon offset, expected %d but got %d" % (
            expected_start_codon_offset, start_codon_offset)

def test_transcript_exons():
    """
    test_transcript_exons : Ensure that properties of CTTNBIP1-004's
    Exon objects match known values.
    """
    transcript = ensembl_grch38.transcript_by_id(CTNNBIP1_004_transcript_id)
    exons = transcript.exons
    assert isinstance(exons, list), \
        "Expected list of Exon objects, got %s : %s" % (exons, type(exons))

    # CTTNBIP1-004 has 5 exons
    assert len(exons) == len(CTTNNIP1_004_exon_lengths), \
        "Expected %d exons but got %d" % (
            len(CTTNNIP1_004_exon_lengths), len(exons))

    for i, exon in enumerate(exons):
        expected_id = CTTNNIP1_004_exon_ids[i]
        assert exon.id == expected_id, \
            "Expected exon #%d of %s to have ID %s but got %s" % (
                i + 1, transcript, expected_id, exon.id)

        expected_length = CTTNNIP1_004_exon_lengths[i]
        assert len(exon) == expected_length, \
            "Expected exon #%d of %s (%s) to have length %d but got %d" % (
                i + 1, transcript, exon, expected_length, len(exon))


# not testing NCBI/Release 54 since I just discovered that ensembl54
# feature='transcript' entries don't have a gene ID.
# TODO: Add gene_id patching to gtf_parsing, add ensembl54 to the list
# below
@test_ensembl_releases(75, 77)
def test_sequence_parts(ensembl):
    """
    test_sequence_parts : Ensure that the UTRs and coding sequence can be
    combined to make the full transcript.
    """
    transcript = ensembl.transcript_by_id(FOXP3_001_transcript_id)

    # The combined lengths of the upstream untranslated region,
    # coding sequence, and downstream untranslated region
    full_sequence = transcript.sequence
    assert_greater(len(full_sequence), 0)

    utr5 = transcript.five_prime_utr_sequence
    assert_greater(len(utr5), 0)

    cds = transcript.coding_sequence
    assert_greater(len(cds), 0)

    utr3 = transcript.three_prime_utr_sequence
    assert_greater(len(utr3), 0)

    # need to use `seq` property of Sequence objects to get underlying
    # strings which can be concatenated and compared
    combined_string = utr5 + cds + utr3

    combined_sequence_length = len(combined_string)
    # make sure length property of transcript matches the sequence length
    eq_(
        combined_sequence_length,
        len(transcript),
        "Length 5' UTR(%dnt) + CDS(%dnt) + 3' UTR(%d) = %d, expected %d" % (
            len(utr5),
            len(cds),
            len(utr3),
            combined_sequence_length,
            len(transcript)))
    eq_(
        combined_string,
        full_sequence,
        "Expected FOXP3-001 sequence:\n%s\n\n5' UTR + CDS + 3' UTR:\n%s" % (
            full_sequence,
            combined_string))

def test_transcript_utr5_sequence_CTNNIP1_004():
    transcript = ensembl_grch38.transcript_by_id(CTNNBIP1_004_transcript_id)
    utr5 = transcript.five_prime_utr_sequence
    expected_utr5_length = len(CTNNBIP1_004_UTR5)
    eq_(len(utr5),
        expected_utr5_length,
        "Expected 5' UTR length %d, got %d" % (
            expected_utr5_length, len(utr5)))
    eq_(utr5, CTNNBIP1_004_UTR5)

def test_transcript_utr3_sequence_CTNNIP1_004():
    transcript = ensembl_grch38.transcript_by_id(CTNNBIP1_004_transcript_id)
    utr3 = transcript.three_prime_utr_sequence
    expected_utr3_length = len(CTNNBIP1_004_UTR3)
    eq_(len(utr3),
        expected_utr3_length,
        "Expected 3' UTR length %d, got %d" % (
            expected_utr3_length, len(utr3)))
    eq_(utr3, CTNNBIP1_004_UTR3)

def test_transcript_cds_CTNNIP1_004():
    transcript = ensembl_grch38.transcript_by_id(CTNNBIP1_004_transcript_id)
    cds = transcript.coding_sequence
    expected_cds_length = len(CTNNBIP1_004_CDS)
    eq_(
        len(cds),
        expected_cds_length,
        "Expected CDS length %d, got %d" % (expected_cds_length, len(cds)))
    eq_(cds, CTNNBIP1_004_CDS)

@test_ensembl_releases()
def test_equal_transcripts(ensembl):
    t1 = ensembl.transcripts_by_name("TP53-001")[0]
    # make an identical gene
    t2 = Transcript(t1.id, ensembl)
    eq_(t1, t2)
    eq_(hash(t1), hash(t2))

@test_ensembl_releases()
def test_not_equal_transcripts(release):
    t1 = release.transcripts_by_name("MUC1-001")[0]
    t2 = release.transcripts_by_name("BRCA1-001")[0]
    assert_not_equal(t1, t2)

def test_protein_id():
    transcript = ensembl_grch38.transcripts_by_name("EGFR-001")[0]
    eq_(transcript.protein_id, "ENSP00000275493")

def test_protein_protein_sequence():
    transcript = ensembl_grch38.transcripts_by_name("EGFR-001")[0]
    eq_(transcript.protein_sequence, EGFR_001_protein_sequence)



from __future__ import absolute_import

from os.path import exists
from .test_common import test_ensembl_releases

@test_ensembl_releases()
def gtf_path_endswith_gtf_gz(ensembl):
    path = ensembl.gtf.local_gtf_path()
    assert exists(path)
    assert path.endswith(".gtf.gz")



from __future__ import absolute_import

from pyensembl import EnsemblRelease

from nose.tools import raises


@raises(Exception)
def test_version_too_old_1():
    EnsemblRelease(1)

@raises(Exception)
def test_version_too_old_47():
    EnsemblRelease(47)

@raises(Exception)
def test_version_is_not_numeric():
    EnsemblRelease("wuzzle")

@raises(Exception)
def test_version_is_none():
    EnsemblRelease(None)

def test_int_version():
    for version in range(54, 77):
        EnsemblRelease(version)

def test_str_version():
    for version in range(54, 77):
        EnsemblRelease(str(version))
